from .rhymeUtils import (PermutedPhone, Permutations, permutation_getters,
                         CONSONANTS)


class RhymeTrieNode(object):

    def __init__(self, phone, parent):
        self.children = {}
        self.parent = parent
        self.phone = phone
        self.words = set()

    def insert(self, phones, word):
        '''Insert a list of phones into this node and its children. Returns the final node of the insert.'''
        if not phones:
            self.words.add(word)
            return self
        child_node = self.children.get(phones[0])
        if child_node is None:
            child_node = RhymeTrieNode(phones[0], self)
            self.children[phones[0]] = child_node
        remaining_phones = phones[1:]
        return child_node.insert(remaining_phones, word)

    def contains(self, phones):
        '''Given a list of phones, finds the end node in the trie associated with those phones.
        Returns a RhymeTrieNode or False if there is no end node associated with the given phones'''
        if phones:
            child_node = self.children.get(phones[0])
            if child_node:
                return child_node.search(phones[1:])
        elif self.words:
            return self
        return False

    def search(self, phones):
        '''Given a list of phones, find a node in the trie associated with those phones.
        Returns a RhymeTrieNode or None if there is no node associated with the given phones'''
        if not phones:
            return self
        child_node = self.children.get(phones[0])
        if child_node:
            return child_node.search(phones[1:])
        return None

    def search_permutations(self, phones):
        '''Returns a generator of nodes'''
        if not phones:
            yield self
            return
        yield from self._add_subtract_phones(phones)
        permuted_phones = self._get_permuted_phones(phones[0])
        remaining_phones = phones[1:]
        for phone in permuted_phones:
            child = self.children.get(phone)
            if child:
                yield from child.search_permutations(remaining_phones)

    def assemble(self):
        '''Aggregate all phones up the trie from this node, inclusive. Returns a generator'''
        if self.phone:
            yield self.phone
            yield from self.parent.assemble()

    def count_nodes(self):
        '''Counts the number of nodes in the trie'''
        return 1 + sum(child.count_nodes() for child in self.children.values())

    def count_words(self):
        '''Counts the number of words in the trie'''
        return len(self.words) + sum(child.count_words() for child in self.children.values())

    def get_sub_words(self):
        yield from self.words
        for child in self.children.values():
            yield from child.get_sub_words()

    def _get_permuted_phones(self, phone):
        if isinstance(phone, PermutedPhone):
            getter = permutation_getters.get(phone.permutation)
            yield from getter(phone.phone)
        else:
            yield phone

    def _add_subtract_phones(self, phones):
        if isinstance(phones[0], PermutedPhone):
            phone = phones[0]
            # try all permutations without this phone
            if phone.permutation == Permutations.SUBTRACTIVE:
                yield from self.search_permutations(phones[1:])
            elif phone.permutation == Permutations.ADDITIVE:
                for consonant in CONSONANTS:
                    # try all permutations with this added consonant
                    child = self.children.get(consonant)
                    if child:
                        yield from child.search_permutations(phones)



'''Utils related to rhyming'''
import warnings
from . import IOUtil
from collections import defaultdict
from enum import Enum


def _auto():
    count = 0
    while True:
        yield count
        count += 1


STRESSED_FLAGS = frozenset(('1', '2'))
VOICED_CONSONANTS = frozenset(('B', 'D', 'G', 'V', 'DH', 'Z', 'ZH', 'JH', 'M',
                               'N', 'NG', 'L', 'R'))
AFFRICATE = 'affricate'
FRICATIVE = 'fricative'
VOWEL = 'vowel'

phone_type_dict, type_phone_dict = IOUtil.load_phone_type_dicts()
word_phone_dict = IOUtil.load_word_phone_dict()
type_voiced_phone_dict = defaultdict(lambda: defaultdict(set))


class PermutedPhone(object):

    def __init__(self, phone, permutation=None):
        self.phone = phone
        self.permutation = permutation

    def __repr__(self):
        return self.phone + ' ' + self.permutation.name


class Permutations(Enum):
    ADDITIVE = _auto()
    SUBTRACTIVE = _auto()
    PARTNER = _auto()
    FAMILY = _auto()
    ASSONANCE = _auto()
    CONSONANT = _auto()
    SUBSTITUTION = _auto()


def is_vowel(phone):
    '''
    Given a phone, determine if it is a vowel
    Returns a boolean
    '''
    if isinstance(phone, PermutedPhone):
        phone = phone.phone
    return phone_type_dict.get(phone) == VOWEL


def is_consonant(phone):
    '''
    Determine if a phone is a consonant
    Returns a boolean
    '''
    return not is_vowel(phone)


CONSONANTS = frozenset(x for x in phone_type_dict if is_consonant(x))
VOWELS = frozenset(x for x in phone_type_dict if is_vowel(x))


def is_voiced(phone):
    '''Given a phone, determine if it is voiced
    Returns a boolean'''
    return phone in VOICED_CONSONANTS or is_vowel(phone)


def extract_syllables(phones):
    '''Extract syllable groupings from a list of phones. Syllables are split by
    vowel, including ending consonants. Leading consonants are grouped with
    the following vowel and consonants (eg DOG -> [[D, AH1, G]])
    Returns a list of lists of string phones'''
    syllables = []
    syllable = []
    # keep track of whether or not we have seen an initial vowel
    seen_vowel = False
    for phone in phones:
        if is_vowel(phone):
            if syllable and seen_vowel:
                syllables.append(syllable)
                syllable = []
            seen_vowel = True
        syllable.append(phone)
    syllables.append(syllable)
    return syllables


def count_syllables(word):
    phones = get_phones(word)
    return len(extract_syllables(phones))


def get_last_stressed(syllables):
    '''
    Gets the last stressed syllable of a list of phones, and any unstressed
    syllables following it.
    TODO: import getting a certain number of syllables
    TODO: distinct from getting x num of syllables?
    Returns a list of lists of string phones.
    '''
    if len(syllables) == 1:
        return syllables
    if is_stressed(syllables[-1]):
        return syllables[-1:]
    else:
        return syllables[-2:]


def is_stressed(syllable):
    '''
    Tests if a syllable (list of string phones) is stressed
    Returns a boolean
    '''
    # first syllable may have a leading consonant
    if is_vowel(syllable[0]):
        vowel = syllable[0]
    else:
        vowel = syllable[1]
    return vowel[-1] in STRESSED_FLAGS


def get_consonant_family(consonant):
    '''Given a consonant, get its family (type, voiced) members'''
    family = phone_type_dict[consonant]
    return type_voiced_phone_dict[family][is_voiced(consonant)]


def get_consonant_partners(consonant):
    '''Given a consonant, get its type members'''
    family = phone_type_dict[consonant]
    return type_phone_dict[family]


def get_last_syllables(word, num_sylls=None):
    # TODO: care about stresses?
    word = word.upper()
    phones = word_phone_dict[word]
    syllables = extract_syllables(phones)
    if num_sylls is None:
        syllables = get_last_stressed(syllables)
    else:
        syllables = syllables[-num_sylls:]
    syllables[0] = strip_leading_consonants(syllables[0])
    return syllables


def strip_leading_consonants(phones):
    for i, phone in enumerate(phones):
        if is_vowel(phone):
            return phones[i:]


def get_phones(word):
    return word_phone_dict[word.upper()]


# TODO: move this to IOUtil? But depends on is_voiced fn
for type_, phones in type_phone_dict.items():
    for phone in phones:
        if is_voiced(phone):
            type_voiced_phone_dict[type_][True].add(phone)
        else:
            type_voiced_phone_dict[type_][False].add(phone)

permutation_getters = {
    Permutations.ADDITIVE: lambda x: [x],
    Permutations.SUBTRACTIVE: lambda x: [x],
    Permutations.PARTNER: get_consonant_partners,
    Permutations.FAMILY: get_consonant_family,
    Permutations.ASSONANCE: lambda x: [x],
    Permutations.CONSONANT: lambda _: VOWELS,
    Permutations.SUBSTITUTION: lambda _: CONSONANTS
}



from . import rhymeUtils as ru
from .util import flatten
from .IOUtil import load_word_phone_dict, load_phone_type_dicts
from .rhymeUtils import PermutedPhone, Permutations
from .RhymeTrieNode import RhymeTrieNode
from itertools import groupby

_rt = None


class Phyme(object):
    '''Phyme: a rhyming dictionary for songwriting'''

    def __init__(self):
        self.rhyme_trie = load_rhyme_trie()

    def search(self, phones):
        '''Search the rhyme trie for sub words given a listen of phones
        Returns a set of strings'''
        result = self.rhyme_trie.search(phones[::-1])
        if result:
            sorted_results = sorted(result.get_sub_words(),
                                    key=ru.count_syllables)
            grouped_results = groupby(sorted_results, key=ru.count_syllables)
            return dict((k, list(v)) for k, v in grouped_results)
        else:
            return None

    def search_permutations(self, phones):
        phones = list(phones)
        nodes = self.rhyme_trie.search_permutations(phones[::-1])
        result_set = set()
        for node in nodes:
            result = node.get_sub_words()
            result_set.update(result)
        sorted_results = sorted(result_set,
                                key=ru.count_syllables)
        grouped_results = groupby(sorted_results, key=ru.count_syllables)
        return dict((k, list(v)) for k, v in grouped_results)

    def get_perfect_rhymes(self, word, num_syllables=None):
        """Get perfect rhymes of a word, defaults to last stressed vowel

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        """

        phones = ru.get_last_syllables(word, num_syllables)
        return self.search(list(flatten(phones)))

    def get_family_rhymes(self, word, num_syllables=None):
        '''
        Get words with the same vowel and stress patterns but with consonants
        from the same family (consonants with the same articulation and
        un/voiced) (DOG -> COB)

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        '''
        phones = ru.get_last_syllables(word, num_syllables)
        if not ru.is_consonant(phones[0][-1]):
            return set()
        phones = flatten(phones)
        phones = map(lambda phone: PermutedPhone(phone, Permutations.FAMILY)
                     if ru.is_consonant(phone)
                     else phone,
                     phones)
        return self.search_permutations(phones)

    def get_partner_rhymes(self, word, num_syllables=None):
        '''
        Get words with the same vowel and stress patterns but with partner
        consonants (consonants with the same articulation) (HAWK -> DOG)

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        '''
        phones = ru.get_last_syllables(word, num_syllables)
        if not ru.is_consonant(phones[0][-1]):
            return set()
        phones = flatten(phones)
        phones = map(lambda phone: PermutedPhone(phone, Permutations.PARTNER)
                     if ru.is_consonant(phone)
                     else phone,
                     phones)
        return self.search_permutations(phones)

    def get_additive_rhymes(self, word, num_syllables=None):
        '''
        Get words with the same vowel and stress patterns but including
        additional consonants (MATTER -> MASTER)

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        '''
        phones = ru.get_last_syllables(word, num_syllables)
        phones = flatten(phones)
        phones = map(lambda phone: PermutedPhone(phone, Permutations.ADDITIVE),
                     phones)
        return self.search_permutations(phones)

    def get_subtractive_rhymes(self, word, num_syllables=None):
        '''
        Get words with the same vowel and stress patterns but dropping some
        consonants (MASTER -> MATTER)

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        '''
        phones = ru.get_last_syllables(word, num_syllables)
        if not ru.is_consonant(phones[0][-1]):
            return set()
        results = set()
        phones = flatten(phones)
        phones = map(lambda phone: PermutedPhone(phone, Permutations.SUBTRACTIVE)
                     if ru.is_consonant(phone)
                     else phone,
                     phones)
        return self.search_permutations(phones)

    def get_consonant_rhymes(self, word, num_syllables=None):
        '''
        Get words with the same stress patterns and consonants but with
        arbitrary vowels (DOG -> BAG)

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        '''
        phones = ru.get_last_syllables(word, num_syllables)
        phones = flatten(phones)
        phones = map(lambda phone: PermutedPhone(phone, Permutations.CONSONANT)
                     if ru.is_vowel(phone)
                     else phone,
                     phones)
        return self.search_permutations(phones)

    def get_assonance_rhymes(self, word, num_syllables=None):
        '''
        Get words with the same vowels and stress patterns but arbitrary
        consonants (JAUNT -> DOG)

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        '''
        phones = ru.get_last_syllables(word, num_syllables)
        phones = flatten(phones)
        phones = map(lambda phone: PermutedPhone(phone, Permutations.ADDITIVE),
                     filter(lambda phone: ru.is_vowel(phone), phones))
        return self.search_permutations(phones)

    def get_substitution_rhymes(self, word, num_syllables=None):
        '''
        Get words with the same vowels and stress patterns but substitute
        arbitrary consonants (FASTER -> FACTOR)

        Arguments:
            word {str} -- word to rhyme

        Keyword Arguments:
            num_syllables {int | None} -- Number of syllables to check
                subtractive rhymes for (default: {None}) for last stressed and
                unstressed

        Returns:
            [set] -- set of rhymes
        '''
        phones = ru.get_last_syllables(word, num_syllables)
        phones = flatten(phones)
        phones = map(lambda phone: PermutedPhone(phone, Permutations.SUBSTITUTION)
                     if ru.is_consonant(phone)
                     else phone,
                     phones)
        return self.search_permutations(phones)


def load_rhyme_trie():
    global _rt
    if _rt:
        return _rt
    word_phone_dict = load_word_phone_dict()
    _rt = RhymeTrieNode(None, None)
    for word, phones in word_phone_dict.items():
        _rt.insert(phones[::-1], word)
    return _rt



'''Utils for loading and parsing pronunciation data into data structures'''
import json
import pickle
import os
from collections import defaultdict

file_path = os.path.dirname(__file__)

# use global objects so we don't load data each time we call a method
word_phone_dict = {}
phone_type_dict = {}
type_phone_dict = {}
type_voiced_phone_dict = defaultdict(lambda: defaultdict(set))
_rt = None


def load_word_phone_dict():
    '''Load a dict of word -> phones mappings'''
    if word_phone_dict:
        return word_phone_dict
    with open(os.path.join(file_path, 'data/word_phone.json')) as f:
        word_phone_dict.update(json.load(f))
    return word_phone_dict


def load_phone_type_dicts():
    '''Load both phone -> type and type -> phone mapped dicts'''
    if phone_type_dict and type_phone_dict:
        return phone_type_dict, type_phone_dict
    with open(os.path.join(file_path, 'data/phone_type.json')) as f:
        phone_type_dict.update(json.load(f))
    with open(os.path.join(file_path, 'data/type_phone.json')) as f:
        type_phone_dict.update(json.load(f))
    return phone_type_dict, type_phone_dict



#!/usr/bin/env python

from setuptools import setup
from Phyme import version

setup(name='Phyme',
      packages=['Phyme'],
      include_package_data=True,
      version=version,
      description='Python rhyming dictionary for songwriting',
      author='James Wenzel',
      author_email='jameswenzel@berkeley.edu',
      url='https://github.com/jameswenzel/Phyme',
      download_url=('https://github.com/jameswenzel/Phyme/tarball/' + version),
      license='MIT License',
      keywords=['songwriting', 'rhyme', 'rhyming', 'dictionary', 'slant',
                'rhymes', 'song', 'writing'],
      classifiers=[])
